import base64
import gzip
import hashlib
import tarfile
import glob
import io
import os
import fnmatch
from itertools import chain

# 1. download the package
# 2. untar it to dest directory with the packagename_version/
# 3. open and read manifest.yaml
# 4. interate on deploy and download if not present (packagename_version/deps/deppackagename_version.kub) and extract
#    the packages to packagename_version/deps/deppackagename_version/.
# 5. interate on deploy and load manifest.yml of all packages
# 6. foreach manifest create the files to packagename_version/resources
# 7. foreach manifest check if resources exists are create it
#

AUTHORIZED_FILES = ["*.libjsonnet",
                    "*.jsonnet",
                    "*.yaml",
                    "README.md",
                    "LICENSE",
                    "AUTHORS",
                    "NOTICE",
                    "manifests",
                    "deps/*.kub"]


AUTHORIZED_TEMPLATES = ["*.yaml",
                        "*.jsonnet",
                        "*.libjsonnet",
                        "*.yml",
                        "*.j2"]


def authorized_files():
    files = []
    for name in AUTHORIZED_FILES:
        for f in glob.glob(name):
            files.append(f)
    for root, _, filenames in os.walk('templates'):
        for name in AUTHORIZED_TEMPLATES:
            for filename in fnmatch.filter(filenames, name):
                files.append(os.path.join(root, filename))
    return files


def all_files():
    files = []
    ignore_files = []
    for filename in ['.helmignore', '.cnrignore', '.kpmignore']:
        if os.path.exists(filename):
            with open(filename, 'r') as f:
                ignore_files.append(f.read().split("\n"))
    ignore_files = list(chain(ignore_files))
    for root, _, filenames in os.walk('.'):
        for filename in filenames:
            if filename in ignore_files:
                continue
            files.append(os.path.join(root, filename))
    return files


def pack_kub(kub, filter_files=True, prefix=None):
    tar = tarfile.open(kub, "w:gz")
    if filter_files:
        files = authorized_files()
    else:
        files = all_files()
    for filepath in files:
        arcname = None
        if prefix:
            arcname = os.path.join(prefix, filepath.replace("./", ""))
        tar.add(filepath, arcname=arcname)

    tar.close()


def unpack_kub(kub, dest="."):
    tar = tarfile.open(kub, "r:gz")
    tar.extractall(dest)
    tar.close()


class CnrPackage(object):
    def __init__(self, blob=None, b64_encoded=True):
        self.files = {}
        self.tar = None
        self.blob = None
        self.io_file = None
        self._digest = None
        self.b64blob = None
        if blob is not None:
            self.load(blob, b64_encoded)

    def _load_blob(self, blob, b64_encoded):
        if b64_encoded:
            self.b64blob = blob
            self.blob = base64.b64decode(blob)
        else:
            self.b64blob = base64.b64encode(blob)
            self.blob = blob

    def load(self, blob, b64_encoded=True):
        self._digest = None
        self._load_blob(blob, b64_encoded)
        self.io_file = io.BytesIO(self.blob)
        self.tar = tarfile.open(fileobj=self.io_file, mode='r:gz')
        for m in self.tar.getmembers():
            tf = self.tar.extractfile(m)
            if tf is not None:
                self.files[tf.name] = tf.read()

    def extract(self, dest):
        self.tar.extractall(dest)

    def pack(self, dest):
        f = open(dest, "wb")
        f.write(self.blob)
        f.close()

    def tree(self, directory=None):
        files = self.files.keys()
        files.sort()
        if directory is not None:
            filtered = [x for x in files if x.startswith(directory)]
        else:
            filtered = files
        return filtered

    def file(self, filename):
        return self.files[filename]

    def isjsonnet(self):
        if "manifest.yaml" in self.files:
            return False
        elif "manifest.jsonnet" in self.files:
            return True
        else:
            raise RuntimeError("Unknown manifest format")

    @property
    def manifest(self):
        manifests_files = ["manifest.yaml", "manifest.jsonnet", "Chart.yaml", "Chart.yml"]
        for filename in manifests_files:
            if filename in self.files:
                return self.files[filename]
        raise RuntimeError("Unknown manifest format")

    @property
    def digest(self):
        if self._digest is None:
            self.io_file.seek(0)
            gunzip = gzip.GzipFile(fileobj=self.io_file, mode='r').read()
            self._digest = hashlib.sha256(gunzip).hexdigest()
        return self._digest



import os
import shutil
from cnrclient.commands.command_base import CommandBase
from cnrclient.utils import mkdir_p


LOCAL_DIR = os.path.dirname(__file__)


class PluginsCmd(CommandBase):
    name = 'plugins'
    help_message = "Install plugins"

    def __init__(self, options):
        super(PluginsCmd, self).__init__(options)
        self.plugin = options.plugin
        self.status = ""

    @classmethod
    def _add_arguments(cls, parser):
        parser.add_argument("plugin", nargs='+',
                            help='plugin cmd')

    def _helm(self):
        default_path = os.getenv("HELM_HOME", "~/.helm")
        home = os.path.expanduser(default_path) + "/plugins/cnr"
        mkdir_p(home)
        source = os.path.join(LOCAL_DIR, "plugins/helm")
        files = [os.path.join(source, "plugin.yaml"),
                 os.path.join(source, "cnr.sh")]

        for f in files:
            shutil.copy(f, home)

        self.status = "helm plugin installed"

    def _call(self):
        if " ".join(self.plugin) == "install helm":
            self._helm()

    def _render_dict(self):
        return {"status": self.status}

    def _render_console(self):
        return self.status



from cnrclient.display import print_packages
from cnrclient.commands.command_base import CommandBase


class ListPackageCmd(CommandBase):
    name = 'list'
    help_message = "list packages"

    def __init__(self, options):
        super(ListPackageCmd, self).__init__(options)
        self.registry_host = options.registry_host
        self.user = options.user
        self.organization = options.organization
        self.query = options.search
        self.media_type = options.media_type
        self.result = None

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_arg(parser)
        cls._add_mediatype_option(parser, default=None, required=False)
        parser.add_argument("-u", "--user", default=None,
                            help="list packages owned by USER")
        parser.add_argument("-o", "--organization", default=None,
                            help="list ORGANIZATION packages")
        parser.add_argument("-s", "--search", default=None,
                            help="search query")

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        params = {}
        if self.user:
            params['username'] = self.user
        if self.organization:
            params["namespace"] = self.organization
        if self.query:
            params['query'] = self.query
        if self.media_type:
            params['media_type'] = self.media_type

        self.result = client.list_packages(params)

    def _render_dict(self):
        return self.result

    def _render_console(self):
        return print_packages(self.result)



import re
import json
import logging
import requests
from requests.utils import urlparse
import cnrclient
from cnrclient.discovery import ishosted, discover_sources
from cnrclient.auth import CnrAuth
from cnrclient.config import CnrConfig


logger = logging.getLogger(__name__)
DEFAULT_REGISTRY = 'http://localhost:5000'


class CnrClient(object):
    def __init__(self, endpoint=DEFAULT_REGISTRY,
                 auth=None, config=None, insecure=False):
        if not auth:
            auth = CnrAuth()
        if not config:
            config = CnrConfig()
        self.auth = auth
        self.config = config
        self.endpoint = self._configure_endpoint(endpoint, insecure)
        self.host = self.endpoint.geturl()
        self._headers = {'Content-Type': 'application/json',
                         'User-Agent': "cnrpy-cli/%s" % cnrclient.__version__}

    def _url(self, path):
        return self.endpoint.geturl() + path

    def auth_token(self):
        """ return the Authorization bearer """
        return self.auth.token(self.host)

    def _configure_endpoint(self, endpoint, insecure):
        if endpoint is None:
            endpoint = DEFAULT_REGISTRY
        alias = self.config.get_registry_alias(endpoint)
        if alias:
            endpoint = alias
        if not re.match("https?://", endpoint):
            if insecure or str.startswith(endpoint, "localhost"):
                scheme = "http://"
            else:
                scheme = "https://"
            endpoint = scheme + endpoint
        return urlparse(endpoint)

    @property
    def headers(self):
        token = self.auth_token()
        headers = {}
        headers.update(self._headers)
        if token is not None:
            headers['Authorization'] = token
        return headers

    def version(self):
        path = "/version"
        resp = requests.get(self._url(path), headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def show_package(self, package, version, media_type=None):
        path = "/api/v1/packages/%s" % (package)
        if version and version != 'default':
            path = path + "/%s" % version
        params = {}
        if media_type:
            params["media_type"] = media_type
        resp = requests.get(self._url(path), params=params, headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def _get_pull_url(self, package, version_parts, media_type):
        if media_type is None:
            raise ValueError("media-type is not set")

        organization, name = package.split("/")
        if version_parts['key'] == "digest":
            digest = version_parts['value']
            url = "/api/v1/packages/%s/%s/blobs/sha256/%s" % (organization, name, digest)
        elif version_parts['key'] == "channel":
            chan_name = version_parts['value']
            channel = self.show_channels(package, chan_name)
            version = channel['current']
            url = "/api/v1/packages/%s/%s/%s/%s/pull" % (organization, name, version, media_type)
        elif version_parts['key'] == "version":
            version = version_parts['value']
            url = "/api/v1/packages/%s/%s/%s/%s/pull" % (organization, name, version, media_type)
        else:
            url = "/api/v1/packages/%s/%s/%s/%s/pull" % (organization, name, version_parts['value'], media_type)
        return url

    def _pull_path(self, name, version_parts, media_type):
        if ishosted(name):
            sources = discover_sources(name, version_parts['value'], media_type)
            path = sources[0]
        else:
            path = self._url(self._get_pull_url(name, version_parts, media_type))
        return path

    def pull(self, name, version_parts, media_type):
        path = self._pull_path(name, version_parts, media_type)
        resp = requests.get(path, headers=self.headers)
        resp.raise_for_status()
        return resp.content

    def pull_json(self, name, version_parts, media_type):
        path = self._pull_path(name, version_parts, media_type)
        resp = requests.get(path, params={'format': 'json'}, headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def list_packages(self, params):
        path = "/api/v1/packages"
        resp = requests.get(self._url(path), params=params, headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def push(self, name, body, force=False):
        organization, pname = name.split("/")
        body['name'] = pname
        body['organization'] = organization
        body['package'] = name
        path = "/api/v1/packages/%s/%s" % (organization, pname)
        resp = requests.post(self._url(path),
                             params={"force": str(force).lower()},
                             data=json.dumps(body), headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def delete_package(self, name, version, media_type):
        organization, name = name.split("/")
        path = "/api/v1/packages/%s/%s/%s/%s" % (organization, name, version, media_type)
        resp = requests.delete(self._url(path), headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def _crud_channel(self, name, channel='', action='get'):
        if channel is None:
            channel = ''
        path = "/api/v1/packages/%s/channels/%s" % (name, channel)
        resp = getattr(requests, action)(self._url(path), params={}, headers=self.headers)
        if channel == '' and resp.status_code == 404:
            return []
        resp.raise_for_status()

        return resp.json()

    def show_channels(self, name, channel=None):
        return self._crud_channel(name, channel)

    def create_channel(self, name, channel):
        return self._crud_channel(name, channel, 'post')

    def delete_channel(self, name, channel):
        return self._crud_channel(name, channel, 'delete')

    def create_channel_release(self, name, channel, release):
        path = "%s/%s" % (channel, release)
        return self._crud_channel(name, path, 'post')

    def delete_channel_release(self, name, channel, release):
        path = "%s/%s" % (channel, release)
        return self._crud_channel(name, path, 'delete')

    def login(self, username, password):
        path = "/api/v1/users/login"
        resp = requests.post(self._url(path),
                             data=json.dumps({"user": {"username": username, "password": password}}),
                             headers=self.headers)
        resp.raise_for_status()
        result = resp.json()
        self.auth.add_token(self.host, result['token'])
        return result

    def signup(self, username, password, password_confirmation, email):
        path = "/api/v1/users"
        resp = requests.post(self._url(path),
                             data=json.dumps({"user": {"username": username,
                                                       "password": password,
                                                       "password_confirmation": password_confirmation,
                                                       "email": email}}),
                             headers=self.headers)
        resp.raise_for_status()
        result = resp.json()
        self.auth.add_token(self.host, result['token'])
        return result



from tabulate import tabulate

from cnrclient.utils import get_media_type


def print_packages(packages):
    header = ['app', 'release', 'downloads', 'manifests']
    table = []
    for package in packages:
        release = package["default"]
        manifests = ", ".join(package['manifests'])
        table.append([package['name'], release, str(package.get('downloads', '-')), manifests])
    return tabulate(table, header)


def print_package_info(packages, extended=False):
    header = ["release", "manifest", "digest"]
    if extended:
        header = header + ["created_at", "size", "downloads"]
    table = []
    for package in packages:
        row = [
            package['release'],
            get_media_type(package['content']['mediaType']),
            package['content']['digest'],
            ]
        if extended:
            row = row + [
                package['created_at'],
                package['content']['size'],
                package.get('downloads', '-'),
            ]
        table.append(row)
    return tabulate(table, header)


def print_channels(channels):
    header = ['channel', 'release']
    table = []
    for channel in channels:
        table.append([channel['name'], channel['current']])
    return tabulate(table, header)



import os
from cnrclient.pack import CnrPackage
from cnrclient.commands.command_base import CommandBase


class PullCmd(CommandBase):
    name = 'pull'
    help_message = "download a package"

    def __init__(self, options):
        super(PullCmd, self).__init__(options)
        self.package = options.package
        self.registry_host = options.registry_host
        self.version = options.version
        self.version_parts = options.version_parts
        self.dest = options.dest
        self.media_type = options.media_type
        self.tarball = options.tarball
        self.path = None

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_mediatype_option(parser)
        cls._add_packagename_option(parser)
        cls._add_packageversion_option(parser)
        parser.add_argument("--dest", default="./",
                            help="directory used to extract resources")
        parser.add_argument("--tarball", action="store_true", default=False,
                            help="download the tar.gz")

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        pullpack = client.pull_json(self.package, version_parts=self.version_parts, media_type=self.media_type)
        package = CnrPackage(pullpack['blob'], b64_encoded=True)
        filename = pullpack['filename']
        self.path = os.path.join(self.dest, filename)
        if self.tarball:
            with open(self.path, 'wb') as tarfile:
                tarfile.write(package.blob)
        else:
            self.path = self.path.split(".tar.gz")[0]
            package.extract(self.path)

    def _render_dict(self):
        return {"pull": self.package,
                "media_type": self.media_type,
                "version": self.version,
                "path": self.path}

    def _render_console(self):
        return "Pull package: %s... \n%s" % (self.package, self.path)



from cnrclient.commands.command_base import CommandBase
from cnrclient.display import print_channels


class ChannelCmd(CommandBase):
    name = 'channel'
    help_message = "Manage package channels"

    def __init__(self, options):
        super(ChannelCmd, self).__init__(options)
        self.package = options.package
        self.registry_host = options.registry_host
        self.delete = options.delete
        self.channel = options.channel
        self.remove = options.remove_release
        self.add = options.set_release
        self.version = options.version
        self.version_parts = options.version_parts
        self.status = None
        self.channels = {}

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_packagename_option(parser)
        cls._add_packageversion_option(parser)

        parser.add_argument("-c", "--channel", default=None,
                            help="channel name")
        parser.add_argument("--set-release", default=False, action='store_true',
                            help="Add release to the channel")
        parser.add_argument("--delete", default=False, action='store_true',
                            help="delete the channel")
        parser.add_argument("--remove-release", default=False, action='store_true',
                            help="Remove a release from the channel")

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        package = self.package
        name = self.channel
        if self.delete is True:
            self.channels = client.delete_channel(package, name)
            self.status = ">>> Channel '%s' on '%s' deleted" % (name, package)
        elif self.add:
            self.channels = client.create_channel_release(package, name, self.version)
            self.status = ">>> Release '%s' added on '%s'" % (self.version, name)
        elif self.remove:
            self.channels = client.delete_channel_release(package, name, self.version)
            self.status = ">>> Release '%s' removed from '%s'" % (self.version, name)
        else:
            self.channels = client.show_channels(package, name)
            if name is not None:
                self.channels = [self.channels]
            self.status = print_channels(self.channels)

    def _render_dict(self):
        return self.channels

    def _render_console(self):
        return "%s" % self.status



import os

from cnrclient.commands.command_base import CommandBase
from cnrclient.config import CnrConfig


LOCAL_DIR = os.path.dirname(__file__)


class ConfigCmd(CommandBase):
    name = 'config'
    help_message = "Install config"

    def __init__(self, options):
        super(ConfigCmd, self).__init__(options)
        self.plugin = options.plugin
        self.status = {}

    @classmethod
    def add_alias(cls, options):
        alias = options.alias
        target = options.host
        config = CnrConfig()
        config.add_registry_alias(alias, target)

    @classmethod
    def _add_arguments(cls, parser):
        sub = parser.add_subparsers()
        command = sub.add_parser('alias')
        command.add_argument("alias", help="Registry alias name")
        command.add_argument("host", help="url")
        command.set_defaults(func=cls.add_alias)

    def _call(self):
        pass

    def _render_dict(self):
        return self.status

    def _render_console(self):
        return self._render_yaml()



from cnrclient.commands.command_base import CommandBase
from cnrclient.auth import CnrAuth


class LogoutCmd(CommandBase):
    name = 'logout'
    help_message = "logout"

    def __init__(self, options):
        super(LogoutCmd, self).__init__(options)
        self.status = None
        self.registry_host = options.registry_host

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_arg(parser)

    def _call(self):
        CnrAuth().delete_token(self.registry_host)
        self.status = "Logout complete"
        if self.registry_host != '*':
            self.status += " from %s" % self.registry_host

    def _render_dict(self):
        return {"status": self.status, 'host': self.registry_host}

    def _render_console(self):
        return " >>> %s" % self.status



from cnrclient.formats.helm.manifest_chart import ManifestChart
from cnrclient.formats.kub_base import KubBase


class Chart(KubBase):
    media_type = "helm"
    platform = "helm"
    manifest_file = ['Chart.yml', 'Chart.yaml']

    def _create_manifes(self):
        return ManifestChart(self.package)

    @property
    def author(self):
        return self.manifest.package['author']

    @property
    def version(self):
        return self.manifest.package['version']

    @property
    def description(self):
        return self.manifest.package['description']

    @property
    def name(self):
        return self.manifest.package['name']

    @property
    def resources(self):
        return self.package.files



import os
from cnrclient.formats.helm.chart import Chart
from cnrclient.formats.kpm.kpm import Kpm


kub_formats = [Chart, Kpm]
kub_by_name = {k.media_type: k for k in kub_formats}
kub_by_platforms = {k.platform: k for k in kub_formats}


def kub_factory(name, *args, **kwargs):
    kub_class = kub_by_name[name]
    target = kwargs.pop('convert_to', None)
    k = kub_class(*args, **kwargs)
    if target is not None and target != kub_class.target:
        k = k.convert_to(target)
    return k


def detect_format(path="."):
    for kub_class in kub_formats:
        for filename in kub_class.manifest_file:
            if os.path.exists(os.path.join(path, filename)):
                return kub_class
    raise ValueError("Unknown manifest format")



from cnrclient.display import print_package_info
from cnrclient.commands.command_base import CommandBase


class ShowCmd(CommandBase):
    name = 'show'
    help_message = "print the package manifest"

    def __init__(self, options):
        super(ShowCmd, self).__init__(options)
        self.package = options.package
        self.registry_host = options.registry_host
        self.version = options.version
        self.verbose = options.wide
        self.media_type = options.media_type
        self.result = None

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_packagename_option(parser)
        cls._add_packageversion_option(parser)
        cls._add_mediatype_option(parser, default=None, required=False)
        parser.add_argument("-w", "--wide",
                            help="Extend display informations", action="store_true", default=False)

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        self.result = client.show_package(self.package, version=self.version, media_type=self.media_type)

    def _render_dict(self):
        return self.result

    def _render_console(self):
        return "Info: %s\n\n" % self.package + print_package_info(self.result, self.verbose)



from cnrclient.pack import CnrPackage
from cnrclient.commands.command_base import CommandBase


class InspectCmd(CommandBase):
    name = 'inspect'
    help_message = "Browse package files"

    def __init__(self, options):
        super(InspectCmd, self).__init__(options)
        self.package = options.package
        self.registry_host = options.registry_host
        self.version = options.version
        self.version_parts = options.version_parts
        self.file = options.file
        self.tree = options.tree
        self.media_type = options.media_type
        self.result = None
        self.format = options.media_type

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_mediatype_option(parser)
        cls._add_packagename_option(parser)
        cls._add_packageversion_option(parser)
        parser.add_argument('--tree', help="List files inside the package", action='store_true', default=True)
        parser.add_argument('-f', '--file', help="Display a file", default=None)

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        result = client.pull(self.package, version_parts=self.version_parts, media_type=self.media_type)
        package = CnrPackage(result, b64_encoded=False)
        if self.file:
            self.result = package.file(self.file)
        elif self.tree:
            self.result = "\n".join(package.tree())
        else:
            self.result = package.manifest

    def _render_dict(self):
        return {"inspect": self.package,
                "output": self.result}

    def _render_console(self):
        return self.result



import argparse
import os
import base64

import requests

from cnrclient.pack import pack_kub
from cnrclient.utils import package_filename
from cnrclient.commands.command_base import CommandBase
from cnrclient.formats.helm.manifest_chart import ManifestChart
from cnrclient.formats import detect_format


class PushCmd(CommandBase):
    name = 'push'
    help_message = "push a package to the registry"
    default_media_type = None

    def __init__(self, options):
        super(PushCmd, self).__init__(options)
        self.registry_host = options.registry_host
        self.force = options.force
        self.manifest = None
        self.namespace = options.ns
        self.media_type = options.media_type
        self.channel = options.channel
        self.version = options.version
        self.package_name = options.name
        self.filter_files = True
        self.metadata = None
        self.prefix = None
        self.pname = None
        self.status = ''

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_arg(parser)
        cls._add_mediatype_option(parser, cls.default_media_type, required=False)
        cls._add_packageversion_option(parser)
        parser.add_argument("-f", "--force", action='store_true', default=False,
                            help="force push")
        parser.add_argument("--ns", "--namespace", default=None,
                            help="package namespace")
        parser.add_argument("--name", default=None,
                            help=argparse.SUPPRESS)
        parser.add_argument("-c", "--channel", default=None,
                            help="Set a channel")

    def _push(self):
        client = self.RegistryClient(self.registry_host)
        filename = package_filename(self.package_name, self.version, self.media_type)
        # @TODO: Pack in memory
        kubepath = os.path.join(".", filename + ".tar.gz")
        pack_kub(kubepath, filter_files=self.filter_files, prefix=self.prefix)
        kubefile = open(kubepath, 'rb')
        body = {"name": self.package_name,
                "release": self.version,
                "metadata": self.metadata,
                "media_type": self.media_type,
                "blob": base64.b64encode(kubefile.read())}
        try:
            client.push(self.package_name, body, self.force)
            self.status = "package: %s (%s | %s) pushed\n" % (self.package_name, self.version, self.media_type)
        except requests.exceptions.RequestException as exc:
            if not (self.channel and exc.response.status_code in [409, 400]):
                raise
        kubefile.close()
        os.remove(kubepath)
        if self.channel:
            client.create_channel_release(self.package_name, self.channel, self.version)
            self.status += ">>> Release '%s' added to '%s'" % (self.version, self.channel)

    def _chart(self):
        self.manifest = ManifestChart()
        if self.pname is None:
            self.pname = self.manifest.name
        self.prefix = self.pname
        self.filter_files = False
        if self.namespace is None:
            raise argparse.ArgumentTypeError("Missing option: --namespace")
        self.package_name = "%s/%s" % (self.namespace, self.pname)
        self.version = self.manifest.version
        self.metadata = self.manifest.metadata()

    def _all_formats(self):
        self.filter_files = False
        if self.version is None or self.version == "default":
            raise argparse.ArgumentTypeError("Missing option: --version")
        if self.package_name is None:
            raise argparse.ArgumentTypeError("Missing option: --name")
        self.namespace, self.pname = self.package_name.split("/")

    def _kpm(self):
        raise NotImplementedError

    def _init(self):
        if self.media_type is None:
            self.media_type = detect_format(".").media_type
        if self.media_type in ["kpm", "kpm-compose"]:
            self._kpm()
        if self.media_type in ["docker-compose"]:
            self._all_formats()
        elif self.media_type in ['helm', 'chart']:
            self._chart()

    def _call(self):
        self._init()
        self._push()

    def _render_dict(self):
        return {"package": self.package_name,
                "version": self.version,
                "media_type": self.media_type,
                "channel": self.channel}

    def _render_console(self):
        return self.status



from builtins import input

import getpass
import argparse

from cnrclient.commands.command_base import CommandBase


class LoginCmd(CommandBase):
    name = 'login'
    help_message = "login"

    def __init__(self, options):
        super(LoginCmd, self).__init__(options)
        self.registry_host = options.registry_host
        self.signup = options.signup
        self.password = options.password
        self.email = options.email
        self.user = options.user
        self.status = None

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_arg(parser)
        parser.add_argument("-s", "--signup", action='store_true', default=False,
                            help="Create a new account and login")
        parser.add_argument("-u", "--user", nargs="?", default=None,
                            help="username")
        parser.add_argument("-p", "--password", nargs="?", default=None,
                            help="password")
        parser.add_argument("-e", "--email", nargs="?", default=None,
                            help="email for signup")

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        if self.user is not None:
            user = self.user
        else:
            user = input("Username: ")
        if self.password is not None:
            p1 = self.password
        else:
            p1 = getpass.getpass()

        if self.signup:
            if self.password is not None:
                p2 = p1
            else:
                p2 = getpass.getpass('Password confirmation: ')
            if self.email is not None:
                email = self.email
            else:
                email = input("Email: ")
            if p1 != p2:
                raise argparse.ArgumentError("password", message="Error: password mismatch")
            client.signup(user, p1, p2, email)
            self.status = "Registration complete"
        else:
            client.login(user, p1)
            self.status = "Login succeeded"

    def _render_dict(self):
        return {"user": self.user, "status": self.status, "host": self.registry_host}

    def _render_console(self):
        return " >>> %s" % self.status



import os
import errno
import re


PACKAGE_REGEXP = r"^(.*?\/)?([a-z0-9_-]+\/[a-z0-9_-]+)([:@][a-z0-9._-]+|@sha256:[a-z0-9]+)?$"
# r"^(.*?)?\/?([a-z0-9_-]+\/[a-z0-9_-]+?)([:@].*)?$"


def get_media_type(mediatype):
    if mediatype:
        match = re.match(r"application/vnd\.cnr\.[a-z_-]+\.(.+?)\.(.+).(.+)", mediatype)
        if match:
            mediatype = match.group(1)
    return mediatype


def package_filename(name, version, media_type):
    return "%s_%s_%s" % (name.replace("/", "_"), version, media_type)


def parse_version(version):
    if version is None or version == "default":
        return {'key': "version", "value": "default"}
    elif str.startswith(version, "@sha256:"):
        return {'key': 'digest',
                'value': version.split("@sha256:")[1]}
    elif version[0] == "@":
        return {'key': 'version',
                'value': version[1:]}
    elif version[0] == ":":
        return {'key': 'channel',
                'value': version[1:]}
    else:
        return {'key': 'unknown', 'value': version}


def split_package_name(name):
    sp = name.split("/")
    package_parts = {"host": None,
                     "namespace": None,
                     "package": None,
                     "version": None}

    if len(sp) >= 1:
        package_parts["host"] = sp[0]
    if len(sp) >= 2:
        package_parts["namespace"] = sp[1]
    if len(sp) >= 3:
        match = re.match(r"^([a-z0-9_-]+?)([:@].*)?$", sp[2])
        package, version = match.groups()
        package_parts['package'] = package
        package_parts['version'] = version
    return package_parts


def parse_package_name(name, regexp=PACKAGE_REGEXP):
    package_regexp = regexp
    match = re.match(package_regexp, name)
    if match is None:
        raise ValueError("Package '%s' does not match format '[registry/]namespace/name[@version|:channel]'" % (name))
    host, package, version = match.groups()
    if not version:
        version = 'default'
    if not host:
        host = None
    else:
        host = host[:-1]
    namespace, package = package.split("/")
    return {'host': host,
            'namespace': namespace,
            'package': package,
            'version': version}


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



from cnrclient.commands.command_base import CommandBase


class DeletePackageCmd(CommandBase):
    name = 'delete-package'
    help_message = 'delete package from the registry'

    def __init__(self, options):
        super(DeletePackageCmd, self).__init__(options)
        self.package = options.package
        self.registry_host = options.registry_host
        self.version = options.version
        self.media_type = options.media_type
        self.result = None

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_mediatype_option(parser)
        cls._add_packagename_option(parser)
        cls._add_packageversion_option(parser)

    def _call(self):
        client = self.RegistryClient(self.registry_host)
        self.result = client.delete_package(self.package, version=self.version, media_type=self.media_type)

    def _render_dict(self):
        return self.result

    def _render_console(self):
        return "Deleted package: %s - %s" % (self.result['package'], self.result['release'])



import argparse
import os

from cnrclient.commands.push import PushCmd
from cnrclient.commands.inspect import InspectCmd
from cnrclient.commands.pull import PullCmd
from cnrclient.commands.version import VersionCmd
from cnrclient.commands.show import ShowCmd
from cnrclient.commands.login import LoginCmd
from cnrclient.commands.logout import LogoutCmd
from cnrclient.commands.channel import ChannelCmd
from cnrclient.commands.list_package import ListPackageCmd
from cnrclient.commands.delete_package import DeletePackageCmd
from cnrclient.commands.plugins import PluginsCmd
from cnrclient.commands.config import ConfigCmd


def all_commands():
    return {
        InspectCmd.name: InspectCmd,
        PushCmd.name: PushCmd,
        VersionCmd.name: VersionCmd,
        PullCmd.name: PullCmd,
        ShowCmd.name: ShowCmd,
        LoginCmd.name: LoginCmd,
        LogoutCmd.name: LogoutCmd,
        ChannelCmd.name: ChannelCmd,
        DeletePackageCmd.name: DeletePackageCmd,
        PluginsCmd.name: PluginsCmd,
        ConfigCmd.name: ConfigCmd,
        ListPackageCmd.name: ListPackageCmd,
    }


def get_parser(commands):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='command help')
    for command_class in commands.values():
        command_class.add_parser(subparsers)
    return parser


def cli():
    try:
        parser = get_parser(all_commands())
        args = parser.parse_args()
        args.func(args)
    except (argparse.ArgumentTypeError, argparse.ArgumentError) as exc:
        if os.getenv("CNR_DEBUG", "true"):
            raise
        else:
            parser.error(exc.message)



import io
import tempfile
import tarfile
import os.path

from requests.utils import urlparse

from cnrclient.client import CnrClient
import cnrclient.pack as packager


class KubBase(object):
    media_type = NotImplementedError
    target = NotImplementedError
    kub_class = NotImplementedError
    manifest_file = []
    cnr_client = CnrClient

    def __init__(self, name,
                 version='default',
                 endpoint=None):
        self._deploy_name = name

        self.endpoint = endpoint
        self._registry = self.cnr_client(endpoint=self.endpoint)
        self._deploy_version = version
        self._package = None
        self._manifest = None

    @property
    def package(self):
        if self._package is None:
            result = self._fetch_package()
            self._package = packager.CnrPackage(result, b64_encoded=False)
        return self._package

    def _create_manifest(self):
        raise NotImplementedError

    @property
    def manifest(self):
        if self._manifest is None:
            self._manifest = self._create_manifest()
        return self._manifest

    def __unicode__(self):
        return ("(<{class_name}({name}=={version})>".format(class_name=self.__class__.__name__,
                                                            name=self.name,
                                                            version=self.version))

    def __str__(self):
        return self.__unicode__().encode('utf-8')

    def __repr__(self):
        return self.__str__()

    @property
    def author(self):
        pass

    @property
    def version(self):
        return self.manifest.version

    @property
    def description(self):
        pass

    @property
    def name(self):
        return self.manifest.name

    @property
    def variables(self):
        pass

    def _fetch_package(self):
        parse = urlparse(self._deploy_name)
        if parse.scheme in ["http", "https"]:
            # @TODO
            pass
        elif parse.scheme == "file":
            parts = parse.path.split("/")
            _, ext = os.path.splitext(parts[-1])
            if ext == ".gz":
                filepath = parse.path
            else:
                filepath = tempfile.NamedTemporaryFile().name
                packager.pack_kub(filepath)
            with open(filepath, "rb") as tarf:
                return tarf.read()
        else:
            return self._registry.pull(self._deploy_name, self._deploy_version, self.media_type)

    def make_tarfile(self, source_dir):
        output = io.BytesIO()
        with tarfile.open(fileobj=output, mode="w:gz") as tar:
            tar.add(source_dir, arcname=os.path.basename(source_dir))
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.md') as readme_file:
    readme = readme_file.read()

base_requirements = [
    'future',
    'requests',
]

cli_requirements = [
    'tabulate',
    'termcolor',
    'pyyaml',
    ]

test_requirements = [
    "pytest",
    "coverage",
    "pytest-cov",
    "pytest-ordering",
    "requests-mock"
]

requirements = base_requirements + cli_requirements

setup(
    name='cnr',
    version='0.3.5',
    description="cloud-native app registry server",
    long_description=readme,
    author="Antoine Legrand",
    author_email='2t.antoine@gmail.com',
    url='https://github.com/ant31/cn-app-registry/cnrclient-server',
    packages=[
        'cnrclient',
        'cnrclient.commands',
        'cnrclient.formats',
        'cnrclient.formats.helm',
        'cnrclient.formats.kpm',
    ],
    scripts=[
        'bin/cnr'
    ],
    package_dir={'cnrclient':
                 'cnrclient'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache License version 2",
    zip_safe=False,
    keywords=['cnrclient', 'cnrclientpy', 'kubernetes'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



import os
import os.path

import yaml

from cnrclient.config import DEFAULT_CONF_DIR
from cnrclient.utils import mkdir_p


class CnrAuth(object):
    """ Store Auth object """

    def __init__(self, conf_directory=DEFAULT_CONF_DIR):
        self.conf_directory = conf_directory
        home = os.path.expanduser("~")
        old_path = "%s/%s/auth_token" % (home, conf_directory)
        path = "%s/%s/auths.yaml" % (home, conf_directory)
        mkdir_p(os.path.join(home, conf_directory))
        self.tokenfile = os.path.join(home, path)
        self._tokens = None
        self._retro_compat(old_path)

    def _retro_compat(self, old):
        oldtoken = self._old_token(old)
        if oldtoken:
            if self.tokens is None or '*' not in self.tokens['auths']:
                self.add_token('*', oldtoken)
            os.remove(old)

    def _old_token(self, path):
        if os.path.exists(path):
            with open(path, 'r') as tokenfile:
                return tokenfile.read()
        else:
            return None

    @property
    def tokens(self):
        if self._tokens is None:
            if os.path.exists(self.tokenfile):
                with open(self.tokenfile, 'r') as tokenfile:
                    self._tokens = yaml.load(tokenfile.read())
            else:
                return None
        return self._tokens

    def token(self, host=None):
        if not self.tokens:
            return None
        if host is None or host not in self.tokens['auths']:
            host = '*'
        return self.tokens['auths'].get(host, None)

    def add_token(self, host, value):
        auths = self.tokens
        if auths is None:
            auths = {'auths': {}}
        auths['auths'][host] = value
        self._write_tokens(auths)

    def _write_tokens(self, tokens):
        with open(self.tokenfile, 'w') as tokenfile:
            tokenfile.write(yaml.safe_dump(tokens, indent=2, default_style='"', default_flow_style=False))

    def delete_token(self, host):
        auths = self.tokens
        if not auths or host not in auths['auths']:
            return None
        prev = auths['auths'].pop(host)
        self._write_tokens(auths)
        return prev



from __future__ import print_function
import sys
import os.path
import yaml


__all__ = ['ManifestChart']


MANIFEST_FILES = ["Chart.yaml", "Chart.yml"]


class ManifestChart(dict):
    def __init__(self, package=None, values=None):
        def __init__(self):
            super(ManifestChart, self).__init__()

        self.values = values
        if package is None:
            self._load_from_path()
        else:
            self._load_yaml(package.manifest)

    def _load_yaml(self, yamlstr):
        try:
            self.update(yaml.load(yamlstr))
        except yaml.YAMLError as exc:
            print("Error in configuration file:", sys.stderr)
            if hasattr(exc, 'problem_mark'):
                mark = exc.problem_mark
                print("Error position: (%s:%s)" % (mark.line+1, mark.column+1), sys.stderr)
            raise exc

    def _load_from_path(self):
        for f in MANIFEST_FILES:
            if os.path.exists(f):
                mfile = f
                break
        with open(mfile) as f:
            self._load_yaml(f.read())

    @property
    def keywords(self):
        return self.get("keywords", [])

    @property
    def engine(self):
        return self.get("engine", "gotpl")

    @property
    def home(self):
        return self.get("home", "")

    @property
    def description(self):
        return self.get("description", "")

    @property
    def version(self):
        return self.get("version", "")

    @property
    def maintainers(self):
        return self.get("maintainers", [])

    @property
    def sources(self):
        return self.get("sources", [])

    @property
    def name(self):
        return self.get("name", [])

    def metadata(self):
        return {"maintainers": self.maintainers,
                "source": self.sources}



import requests

import cnrclient
from cnrclient.commands.command_base import CommandBase


class VersionCmd(CommandBase):
    name = 'version'
    help_message = "show versions"

    def __init__(self, options):
        super(VersionCmd, self).__init__(options)
        self.api_version = None
        self.client_version = None
        self.registry_host = options.registry_host

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_arg(parser)

    def _api_version(self):
        api_version = None
        try:
            client = self.RegistryClient(self.registry_host)
            response = client.version()
            api_version = response
        except requests.exceptions.RequestException:
            api_version = ".. Connection error"
        return api_version

    def _cli_version(self):
        return cnrclient.__version__

    def _version(self):
        return {'api-version': self._api_version(),
                "client-version": self._cli_version()}

    def _call(self):
        version = self._version()
        self.api_version = version['api-version']
        self.client_version = version['client-version']

    def _render_dict(self):
        return {"api-version": self.api_version,
                "client-version": self.client_version}

    def _render_console(self):
        return "\n".join(["Api-version: %s" % self.api_version, "Client-version: %s" % self.client_version])



import re
from html.parser import HTMLParser
import requests


package_regexp = "(.+?)/(.+)"


class MetaHTMLParser(HTMLParser):
    def __init__(self, variables):
        self.meta = {}
        self.variables = variables
        HTMLParser.__init__(self)

    def replace_values(self, s):
        for k, v in self.variables.items():
            s = s.replace("{%s}" % k, v)
        return s

    def handle_starttag(self, tag, attrs):
        if tag == "meta":
            d = dict(attrs)
            if 'name' in d and d['name'] == 'cnr-package':
                name, source = d['content'].split(" ")
                name = self.replace_values(name)
                source = self.replace_values(source)
                if name not in self.meta:
                    self.meta[name] = []
                self.meta[name].append(source)


def split_package_name(package):
    m = re.search(package_regexp, package)
    host, name = (m.group(1), m.group(2))
    return (host, name)


def ishosted(package):
    host, _ = split_package_name(package)
    if "." in host or 'localhost' in host:
        return True
    else:
        return False


def discover_sources(package, version, media_type, secure=False):
    schemes = ["https://", "http://"]
    host, name = split_package_name(package)
    for scheme in schemes:
        url = scheme + host
        try:
            r = requests.get(url, params={"cnr-discovery": 1}, timeout=2)
        except (requests.exceptions.Timeout, requests.ConnectionError) as e:
            if scheme == "https://" and not secure:
                continue
            else:
                raise e

        r.raise_for_status()
        variables = {'name': name, 'version': version,
                     "media_type": media_type,
                     "mediatype": media_type}
        p = MetaHTMLParser(variables)
        p.feed(r.content.decode())
        if package in p.meta:
            return p.meta[package]
    return None



import os
import os.path

import yaml

from cnrclient.utils import mkdir_p


DEFAULT_CONF_DIR = os.getenv('CNR_CONF_DIR', ".cnr")


class CnrConfig(object):
    """ Store Auth object """

    def __init__(self, conf_directory=DEFAULT_CONF_DIR):
        self.conf_directory = conf_directory
        home = os.path.expanduser("~")
        mkdir_p(os.path.join(home, conf_directory))
        path = "%s/%s/config.yaml" % (home, conf_directory)
        self.configfile = os.path.join(home, path)
        self._config = None

    @property
    def config(self):
        if self._config is None:
            if os.path.exists(self.configfile):
                with open(self.configfile, 'r') as configfile:
                    self._config = yaml.load(configfile.read())
            else:
                self._config = {}
        return self._config

    def add_key(self, key, value):
        self.config[key] = value
        self._write_config(self.config)

    def _write_config(self, config):
        with open(self.configfile, 'w') as configfile:
            configfile.write(yaml.safe_dump(config, indent=2, default_style='"', default_flow_style=False))

    def add_registry_alias(self, alias, target):
        if 'repositories' not in self.config:
            self.config['repositories'] = {}
        self.config['repositories'][alias] = target
        self._write_config(self.config)

    def get_registry_alias(self, alias):
        if 'repositories' in self.config:
            return self.config['repositories'].get(alias, None)
        return None



from __future__ import print_function
import argparse
import json

import yaml
import requests

import cnrclient
from cnrclient.client import CnrClient
from cnrclient.utils import parse_package_name, parse_version, split_package_name


def _set_package(parser, namespace, dest, package_parts):
    parsed_version = parse_version(package_parts['version'])
    setattr(namespace, "registry_host", package_parts['host'])
    setattr(namespace, 'version', parsed_version['value'])
    setattr(namespace, 'version_parts', parsed_version)
    package = "%s/%s" % (package_parts['namespace'], package_parts['package'])
    setattr(namespace, dest, package)
    setattr(namespace, "package_parts", package_parts)


class PackageName(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        try:
            name = value[0]
            package_parts = parse_package_name(name)
            _set_package(parser, namespace, self.dest, package_parts)
        except ValueError as exc:
            raise parser.error(exc.message)


class PackageSplit(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        name = value
        package_parts = split_package_name(name)
        _set_package(parser, namespace, self.dest, package_parts)


class CommandBase(object):
    name = 'command-base'
    help_message = 'describe the command'
    RegistryClient = CnrClient
    default_media_type = None

    def __init__(self, args_options):
        self.args_options = args_options
        self.output = args_options.output

    def render(self):
        if self.output == 'none':
            return
        elif self.output == 'json':
            self._render_json()
        elif self.output == 'yaml':
            self._render_yaml()
        else:
            print(self._render_console())

    @classmethod
    def call(cls, options):
        try:
            cls(options)()
        except requests.exceptions.RequestException as exc:
            raise argparse.ArgumentTypeError(exc.message)

    def __call__(self):
        self._call()
        self.render()

    @classmethod
    def add_parser(cls, subparsers):
        parser = subparsers.add_parser(cls.name, help=cls.help_message)
        cls._add_output_option(parser)
        cls._add_arguments(parser)
        parser.set_defaults(func=cls.call)

    def _render_json(self):
        print(json.dumps(self._render_dict(), indent=2, separators=(',', ': ')))

    def _render_dict(self):
        raise NotImplementedError

    def _render_console(self):
        raise NotImplementedError

    def _render_yaml(self):
        print(yaml.safe_dump(self._render_dict()))

    def _call(self):
        raise NotImplementedError

    @classmethod
    def _add_arguments(cls, parser):
        raise NotImplementedError

    @classmethod
    def _add_registryhost_option(cls, parser):
        parser.add_argument("-H", "--registry-host", default=cnrclient.client.DEFAULT_REGISTRY,
                            help='registry API url')

    @classmethod
    def _add_output_option(cls, parser):
        parser.add_argument("--output", default="text", choices=['text',
                                                                 'none',
                                                                 'json',
                                                                 'yaml'],
                            help="output format")

    @classmethod
    def _add_mediatype_option(cls, parser, default=None, required=True):
        if default is None:
            default = cls.default_media_type
        if default is not None:
            required = False

        parser.add_argument("-t", "--media-type", default=default, required=required,
                            help='package format: [kpm, kpm-compose, helm]')

    @classmethod
    def _add_packagename_option(cls, parser):
        parser.add_argument('package', nargs=1, default=None,
                            action=PackageName, help="package-name")

    @classmethod
    def _add_packagesplit_option(cls, parser):
        parser.add_argument('package', nargs="?", default=None,
                            action=PackageSplit, help="registry-host.com/namespace/name")

    @classmethod
    def _add_packageversion_option(cls, parser):
        parser.add_argument("-v", "--version",
                            help="package VERSION", default='default')

    @classmethod
    def _add_registryhost_arg(cls, parser):
        parser.add_argument("registry_host", nargs='?',
                            default=cnrclient.client.DEFAULT_REGISTRY,
                            help='registry API url')
